package fr.inria.gemo.util;

import java.net.Inet4Address;
import java.net.UnknownHostException;
import java.security.SecureRandom;

/**
 * <p>
 * This code is strongly inspired from the j-GUID Stateless Session Bean from
 * Steve Woodcock. The following text comes from his post on <a
 * href="http://www.theserverside.com/patterns/thread.jsp?thread_id=4976">The
 * Server Side </a>.
 * </p>
 * <p>
 * EJB GUID generation techniques rely on a database to provide a sequencing
 * mechanism or singleton objects, or a combination of both. This is a GUID
 * generator that does not need a singleton or a database - can be safely pooled
 * on a single machine and deployed in a cluster to provide completely scalable
 * performance.
 * </p>
 * <p>
 * The problem of generating unique IDs can essentially be broken down as
 * uniqueness over space and uniqueness over time which, when combined, produces
 * a globally unique sequence.
 * </p>
 * <p>
 * Taking the UUID and GUID Internet standards draft for the Network Working
 * Group by Paul J. Leach, Microsoft, and Rich Salz, Certco, as a starting point
 * we assume that the GUID be represented as a 36-digit alphanumeric (including
 * hyphens) of the format xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx.
 * </p>
 * <p>
 * The first 20 characters represent the uniqueness over time (the low 32-bits
 * of the time stamp as the first 8, the next 4 as the mid 16 bits of the time
 * stamp, the next 4 as the high value of the time stamp multiplexed with the
 * version, and the next 4 a combination of the clock sequence high -multiplexed
 * with the variant field - and the low bits) Note: The internet guidelines
 * suggest the timestamp as a 60-bit value to a precision of 100ns since
 * 00:00:00.00, 15 October 1582 (the date of Gregorian reform to the Christian
 * calendar) <br>
 * The last 12 characters are a 48-bit node identifier usually implemented as
 * the IEEE 802 address, which gives uniqueness over space. <br>
 * These are combined to produce a unique sequence.
 * </p>
 * <p>
 * Some of the main problems in an EJB implementation of this technique are:
 * <ol>
 * <li>there is no access to the traditional IEEE 802 address.
 * <li>more than one object can exist at the same time and so generate the same
 * time stamp at a granularity of a millisecond (Java's timestamp granularity).
 * <li>a clock sequence or equivalent is used as a way of reading/writing a
 * value to storage (e.g. a database or system file) that can be used in case
 * the clock is set backwards and as a seed for the number sequence. This is
 * even more of a problem when a cluster of machines do not use the same
 * database.
 * <li>singletons are not portable and not recommended in the EJB specs.
 * </ol>
 * </p>
 * <p>
 * Previous solutions have required a database or a singleton object technique
 * to try and overcome these limitations. However, by taking the following
 * format we can address these issues in a simple manner.
 * </p>
 * <p>
 * <b>1) </b> (1-8 hex characters) use the low 32 bits of
 * System.currentTimeMillis(). We could use the recommended date format by
 * adding 12219292800000L to the system time long before grabbing the low 32
 * bits but there doesn't seem much point. This gives us a uniqueness of a
 * millisecond - therefore any clashing object will have to be generated within
 * the same millisecond.
 * </p>
 * <p>
 * But this does not give us a uniqueness in space or complete uniqueness in
 * time.
 * </p>
 * <p>
 * <b>2) </b> (9-16 hex characters) the IP address of the machine as a hex
 * representation of the 32 bit integer underlying IP - gives us a spatial
 * uniqueness to a single machine - guarantees that these characters will be
 * different for machines in a cluster or on a LAN.
 * </p>
 * <p>
 * We now have machine uniqueness per millisecond - but this does not account
 * for multiple objects on a machine, clock sequence being reset or the same
 * object producing multiple GUIDs per millisecond.
 * </p>
 * <p>
 * <b>3) </b> (17-24 hex characters) the hex value of the Stateless Session bean
 * object's hashCode (a 32 bit int) - in the Java language spec - the hashcode
 * value of Object is defined as: <br>
 * <code>      ** As much as is reasonably practical, the hashCode method defined by class
 *      object does return distinct integers for distinct objects. (This is
 *      typically implemented by converting the internal address of the object
 *      into an integer, but this implementation technique is not required by the
 *      Java programming language.)**</code>
 * <br>
 * This removes the need for a singleton object as the Stateless bean uses its
 * own hashCode as part of the GUID - so even if two objects create a value at
 * the same millisecond on the same machine these characters will be different.
 * Remember the hashCode is not over-ridden here so the implementation will be
 * the one for Object. Note - on machines with low RAM the most significant
 * characters will always be 00 in the hashcode sequence.
 * </p>
 * <p>
 * So we now have a number that is unique to the millisecond, the IP address and
 * the object creating it. <br>
 * There is still the possibility that the same object can produce many GUIDs
 * within the same millisecond or the clock may be reset.
 * </p>
 * <p>
 * <b>4) </b> (25-32 hex characters) a random 32 bit integer generated for each
 * method invocation from the SecureRandom java class using
 * SecureRandom.nextInt(). This method produces a cryptographically strong
 * pseudo-random integer. The Java lang defines this as: <br>
 * <code>    ** Returns a pseudo-random, uniformly distributed int value drawn from this
 *    random number generator's sequence. The general contract of nextInt is that
 *    one int value in the specified range is pseudorandomly generated and returned.
 *    All n possible int values are produced with (approximately) equal probability.**</code>
 * </p>
 * <p>
 * By adding a random int to the GUID for each new request we now have a GUID
 * that has a combination of the following four properties
 * <ol>
 * <li>is unique to the millisecond
 * <li>is unique to the machine
 * <li>is unique to the object creating it
 * <li>is unique to the method call for the same object
 * </ol>
 * <br>
 * The only possible theoretical conflicts are:
 * <ol>
 * <li>that two objects on the same machine are assigned the exact same
 * hashCode (I do not know of any implementations that do this but there may be
 * some out there) and at the same millisecond must also get the same integer
 * value from the SecureRandom implementation.
 * <li>The same int value is returned from the SecureRandom object in
 * subsequent method calls for the same object.
 * <li>A reset clock (which would require a redeployment of the bean) will
 * produce an identical hashcode for the new deployment as a previous one AND
 * the random values will have to be the same in the same repeated millisecond
 * value as in a previous sequence.
 * </ol>
 * </p>
 * <p>
 * There is no problem using this approach to create a pool of GUID generators
 * or using such pools in a cluster. Also performance is excellent as the
 * SecureRandom seed can be initialised in the ejbCreate method and the
 * subsequent nextInt() numbers are quick to produce.
 * </p>
 * <p>
 * A demo of this code is available at <a
 * href="http://www.activescript.co.uk">www.activescript.co.uk </a>.
 * </p>
 * 
 * @author $Author: benjello $
 * @version $Id: GUIDGenerator.java,v 1.1 2004/09/23 09:49:43 benjello Exp $
 * 
 * Originally, this class was in the fr.inria.gemo.axml.util package but I
 * thought it could be useful to other Gemo projects so I decided to move it to
 * the fr.inria.gemo.util package.
 * 
 * FIXME: Inet4Address.getLocalHost().getAddress() can throw an
 * UnknownHostException when there's no networking. For an ActiveXML Peer it
 * shouldn't be a problem but it doesn't make so much sense to have an AXML Peer
 * w/o networking. Anyway, when an Exception is thrown, something else should be
 * used to simulate the uniqueness in space. One way would be to use a
 * randomized int. Another way would be to use the AXML Peer URL, except that
 * there's no easy way for the GUIDGenerator to get that kind of information and
 * we then we need to reduce that information to 32 bits. MD5 can't even be used
 * because it's too long. Maybe using another encoding like 64 base could work.
 * 
 * Update: I fixed the previous problem on 09/03/2004. See below in the code.
 */

public class GUIDGenerator {

	/*****************************************************************************
	 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	 * 
	 * Data members.
	 *  * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	 */

	// The randomizer used to generate the last part of the GUID.
	private SecureRandom seeder = null;

	// String variable caching some of the data.
	private String midValue;

	/*****************************************************************************
	 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	 * 
	 * Constructors.
	 *  * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	 */

	/**
	 * Constructs a new GUIDGenerator.
	 */
	public GUIDGenerator() {
		// In order to speed up the GUID creation, the midValue is created
		// in the constructor and cached for further use.

		// Create the randomizer.
		seeder = new SecureRandom();

		// Create the temporary buffer to create the mid value of the GUID.
		StringBuffer tmpBuffer = new StringBuffer();

		// Get the IPv4 address of the localhost.
		// We'll have to find another way when IPv6 will
		// be widely used.
		// Should also find a way around when there's no network and
		// thus no network address available.

		byte[] bytes = null;

		try {
			// Inet4Address.getLocalHost().getAddress() returns the raw IP address
			// of this InetAddress object. The result is in network byte order:
			// the highest order byte of the address is in getAddress()[0].
			// So there's no need to reverse the byte[] before using getInt(byte[]).
			bytes = Inet4Address.getLocalHost().getAddress();
			//System.out.println(Inet4Address.getLocalHost().toString());
		} catch (UnknownHostException UHex) {
			/*
			 * What do we do now? That's a good question ;-) What could we use instead
			 * of the IPv4 address? Let's find out later... In our case, it doesn't
			 * make so much sense to have an ActiveXML Peer w/o networking. One way
			 * could be to use a randomized value, like doing: hexInetAddress =
			 * toHexString(seeder.nextInt(),8); or use some host specific information
			 * to calculate these bytes like: java.runtime.version, os.arch,
			 * os.name... and some user specific information.
			 */
			/*
			 * Update 09/03/2004 by Jerome Baumgarten I thought this would never be a
			 * problem until I got it. I actually got this exception on a
			 * misconfigured laptop running knoppix. Somehow, the system was not able
			 * to return an IP address for localhost. So one quick way to fix this is
			 * to fake it.
			 */
			bytes = new byte[] { 127, 0, 0, 1};
		}
		String hexInetAddress = toHexString(getInt(bytes), 8);
		// Get the hashcode.
		String thisHashCode = toHexString(System.identityHashCode(this), 8);

		// Create the cached midValue
		tmpBuffer.append('-');
		tmpBuffer.append(hexInetAddress.substring(0, 4));
		tmpBuffer.append('-');
		tmpBuffer.append(hexInetAddress.substring(4));
		tmpBuffer.append('-');
		tmpBuffer.append(thisHashCode.substring(0, 4));
		tmpBuffer.append('-');
		tmpBuffer.append(thisHashCode.substring(4));
		midValue = tmpBuffer.toString();

		// Load up the first randomized value
		seeder.nextInt();
	}

	/**
	 * Transforms an array of 4 <code>byte</code> s into an <code>int</code>.
	 * bytes[0] is treated as the high 8 bits of the int result. bytes[3] is
	 * treated as the low 8 bits of the int result. We assume that we have 4 bytes
	 * to build an int. If we have less, that's a bit of a problem and an
	 * ArrayIndexOutOfBoundsException exception should be thrown. If there's more
	 * than 4 bytes, that's no big deal.
	 * 
	 * @param bytes
	 *          the <code>byte</code> array you want to transform into an int.
	 * @return the int corresponding to the <code>byte</code> array.
	 */
	private int getInt(byte[] bytes) {
		int r = 0;
		int shift = 24;
		for (int i = 0; shift >= 0; i++) {
			r += (bytes[i] & 0xff) << shift;
			shift -= 8;
		}
		return r;
	}

	/**
	 * Transforms an <code>int</code> into a <code>String</code> of
	 * hexadecimal codes. The hexadecimal code will be in upper case letters.
	 * 
	 * @param i
	 *          the <code>int</code> value you want to convert into an
	 *          hexadecimal <code>String</code>.
	 * @param l
	 *          the length of the <code>String</code> you want to obtain. If the
	 *          hexadecimal value of the <code>int</code> is smaller than l,
	 *          then leading 0 will be added to the resulting <code>String</code>.
	 * @return a <code>String</code> of the hexadecimal codes corresponding to
	 *         the <code>int</code> argument.
	 */
	private String toHexString(int i, int l) {
		// Using upper cases letters for the hex code slows down the process
		// a little bit.
		String r = Integer.toHexString(i).toUpperCase();
		int lgth = r.length();
		if (l == lgth) {
			return r;
		} else {
			StringBuffer rBuffer = new StringBuffer(l);
			// lgth shouldn't be greater than l.
			// Let's assume that and not even test it.
			// If someone reading this code believes it should
			// be tested anyway, go ahead and feel free to add the test.
			// The following adds leading 0 if it's missing.
			for (int j = 0; j < (l - lgth); j++) {
				rBuffer.append('0');
			}
			rBuffer.append(r);
			return rBuffer.toString();
		}
	}

	/**
	 * Returns a GUID formatted as
	 * <code>xxxxxxxx-xxxxxxxx-xxxxxxxx-xxxxxxxx</code>.
	 * <p>
	 * <b>1) </b> (1-8 hex characters) use the low 32 bits of
	 * System.currentTimeMillis().
	 * </p>
	 * <p>
	 * <b>2) </b> (9-16 hex characters) the IP address of the machine as a hex
	 * representation of the 32 bit integer underlying IP.
	 * </p>
	 * <p>
	 * <b>3) </b> (17-24 hex characters) the hex value of this GUIDGenerator
	 * object's hashCode (a 32 bit int).
	 * </p>
	 * <p>
	 * <b>4) </b> (25-32 hex characters) a random 32 bit integer generated for
	 * each method invocation from the SecureRandom java class using
	 * SecureRandom.nextInt(). This method produces a cryptographically strong
	 * pseudo-random integer.
	 * </p>
	 * 
	 * @return the GUID formatted as
	 *         <code>xxxxxxxx-xxxxxxxx-xxxxxxxx-xxxxxxxx</code>.
	 */
	synchronized public String getGUID() {
		// To store the low 32 bits of the time.
		int timeLow = (int) System.currentTimeMillis() & 0xFFFFFFFF;
		//To get only the low 32 bits.
		return (toHexString(timeLow, 8) + midValue + toHexString(seeder.nextInt(),
				8));
	}

	// TO TEST THIS CLASS
	/*
	static public void main(String[] args) {
		GUIDGenerator guid = new GUIDGenerator();
		int nbGUID = Integer.parseInt(args[0]);
		//System.out.println("Generated GUID:");
		long start = System.currentTimeMillis();
		for (int i = 0; i < nbGUID; i++) {
			//System.out.println(guid.getGUID());
			guid.getGUID();
		}
		long end = System.currentTimeMillis();
		System.out.println("Time (ms) to generate " + args[0] + " GUIDs : "
				+ (end - start));
	}
	*/
}