package fr.inria.gemo.util;

/** 
 * <p>
 * This class describes the configuration used for the pretty printing of a source.
 * You can specify if you want to use tabulations or space characters (including the number) 
 * for indentation. You can also specify what should be used for a new line.
 * </p>
 * @author Jerome Baumgarten jerome.baumgarten@inria.fr
 * @version 1.0
 */
public class PrettyPrinterConfiguration {

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	*
	* Constants.
	*
	* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

	/**
	 * A constant holding a newline string as "\n".
	 */
	static final public String NEWLINE_N = "\n";
	/**
	 * A constant holding a newline string as "\r\n".
	 */
	static final public String NEWLINE_RN = "\r\n";

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	*
	* Data members.
	*
	* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

	/**
	 * Should we indent the source?
	 */
	public boolean indent = true;
	/**
	 * Should the indentation use tabulations?
	 */
	public boolean useTab = false;
	/**
	 * The number of spaces used for the indentation.
	 * This value is used if <code>useTab</code> is <code>false</code>.
	 */
	public short space = 2;
	/**
	 * What should be used for a new line.
	 */
	public String newline = NEWLINE_N;
	/**
	 * The <code>String</code> representing the spaces used for indentation.
	 * For example, if <code>useTab</code> is <code>true</code> then it will
	 * be "\t".
	 */
	protected String spaces = "  ";

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	*
	* Constructors.
	*
	* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

	/**
	 * Constructs a new <code>PrettyPrinterConfiguration</code>.
	 */
	public PrettyPrinterConfiguration() {
	}

	/**
	 * Constructs a new <code>PrettyPrinterConfiguration</code>.
	 * @param indent indicates if indentation should be performed for the pretty printing.
	 */
	public PrettyPrinterConfiguration(boolean indent) {
		this.indent = indent;
	}

	/**
	 * Constructs a new <code>PrettyPrinterConfiguration</code>.
	 * @param useTab indicates if tabulations should be use for the indentation.
	 * @param space represents the number of spaces to use for indentation if not
	 * using tabulation.
	 * @param newline represents the <code>String</code> used to do a new line.
	 */
	public PrettyPrinterConfiguration(
		boolean useTab,
		short space,
		String newline) {
		this.useTab = useTab;
		this.space = space;
		if (useTab)
			this.spaces = "\t";
		else
			this.spaces = createSpaceString(this.space);
		this.newline = newline;
	}

	/**
	 * Creates a <code>String</code> of <code>nb</code> space characters.
	 * @param nb the number of space characters (' ') you want the <code>String</code>
	 * to contain.
	 * @return a <code>String</code> of <code>nb</code> space characters.
	 */
	protected String createSpaceString(short nb) {
		if (nb <= 0)
			return StringUtils.EMPTY_STRING;
		StringBuffer _spaces = new StringBuffer(nb);
		for (short i = 0; i < nb; i++)
			_spaces.append(' ');
		return _spaces.toString();
	}
}
