/*
 * Created on 7 mars 2003
 *
 */
package fr.inria.gemo.util.xml;

import org.w3c.dom.DocumentType;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.w3c.dom.Attr;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;

import java.io.Writer;
import java.io.StringWriter;
import java.io.IOException;

import fr.inria.gemo.util.StringUtils;

/**
 * @author J?r?me Baumgarten
 * @version $Id: DOMSerializer.java,v 1.1 2004/09/23 09:49:43 benjello Exp $
 */
public class DOMSerializer {

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	*
	* Constants.
	*
	* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

	/* Constants for the DOM serialization. */
	/**
	 * The code for <pre>&</pre>.
	 */
	static public final String AMP = "&amp;";
	/**
	 * The code for <pre>'</pre>.
	 */
	static public final String APOS = "&apos;";
	/**
	 * The code for <pre>"</pre>.
	 */
	static public final String QUOT = "&quot;";
	/**
	 * The code for <pre><</pre>.
	 */
	static public final String LT = "&lt;";
	/**
	 * The code for <pe>></pre>.
	 */
	static public final String GT = "&gt;";

	/* Constants for the XML pretty printing. */
	/**
	 * CDATA start tag.
	 */
	static public final String CDATA_START = "<![CDATA[";
	/**
	 * CDATA end tag.
	 */
	static public final String CDATA_END = "]]>";
	/**
	 * COMMENT start tag.
	 */
	static public final String COMMENT_START = "<!--";
	/**
	 * COMMENT end tag.
	 */
	static public final String COMMENT_END = "-->";
	/**
	 * PROCESSING INSTRUCTION start tag.
	 */
	static public final String PI_START = "<?";
	/**
	 * PROCESSING INSTRUCTION end tag.
	 */
	static public final String PI_END = "?>";
	/**
	 * DOCTYPE start tag.
	 */
	static public final String DOCTYPE_START = "<!DOCTYPE";
	/**
	 * DOCTYPE end tag.
	 */
	static public final String DOCTYPE_END = ">";
	/**
	 * DOCTYPE SYSTEM Id
	 */
	static public final String DOCTYPE_SYSTEM_ID = "SYSTEM";
	/**
	 * DOCTYPE PUBLIC Id
	 */
	static public final String DOCTYPE_PUBLIC_ID = "PUBLIC";
	/**
	 * XML header.
	 * So far, it is not possible to easily guess information on the encoding and
	 * the standalone characteristic of an XML document.
	 * This kind of information will be available through the DOM Level 3 API
	 * which as of today is only a Working Draft and I only know of Xerces
	 * providing some implementation of the DOM Level 3 API.
	 * Also the version number should not be hard-coded.
	 */
	static public final String XML_HEADER_START = "<?xml";
	static public final String XML_HEADER_VERSION = "version";
	static public final String XML_HEADER_ENCODING = "encoding";
	static public final String XML_HEADER_STANDALONE = "standalone";
	static public final String XML_HEADER_END = "?>";

	/**
	 * Pretty prints an XML <code>Node</code> to a <code>String</code> object.
	 * It uses a default <code>XMLPrettyPrinterConfiguration</code> object for the
	 * pretty printing. The <code>Node</code> and its children (if any) will be printed out
	 * to a <code>String</code> object.
	 * @param node the <code>Node</code> to pretty print.
	 * @return a <code>String</code> object containing the pretty printing of the
	 * <code>Node</code> object.
	 * @see fr.inria.gemo.util.XMLPrettyPrinterConfiguration
	 */
	//	static public String NodeToString(Node node) {
	//		return NodeToString(node, new XMLPrettyPrinterConfiguration());
	//	}
	static public String Serialize(Node node) {
		return Serialize(node, new OutputFormat());
	}

	static public String Serialize(Node node, OutputFormat outputFormat) {
		try {
			StringWriter stringWriter = new StringWriter();
			Serialize(node, outputFormat, stringWriter);
			return stringWriter.toString();
		} catch (IOException e) {
		}
		return StringUtils.EMPTY_STRING;
	}

	/**
	 * Prett prints an XML <code>Node</code> to a <code>String</code> object.
	 * @param node the <code>Node</code> to pretty print.
	 * @param config the <code>XMLPrettyPrinterConfiguration</code> object used
	 * for the pretty printing.
	 * @return a <code>String</code> object containing the pretty printing of the
	 * <code>Node</code> object.
	 * @see fr.inria.gemo.util.XMLPrettyPrinterConfiguration
	 */
	//	static public String NodeToString(
	//		Node node,
	//		XMLPrettyPrinterConfiguration config) {
	//		StringBuffer buffer = new StringBuffer();
	//		NodeToString(node, config, buffer, 0); // 0 is the indentation level.
	//		return buffer.toString();
	//	}

	static public void Serialize(
		Node node,
		OutputFormat outputFormat,
		Writer writer) throws IOException {
			Serialize(node,outputFormat,writer,0);
	}

	/**
	 * Pretty prints an XML <code>Node</code> to a <code>StringBuffer</code> object.
	 * As the printing is done on the <code>Node</code> and its children (if any), the
	 * <code>StringBuffer</code> buffer is used to accumulate the printing during the
	 * on-going process.
	 * @param node the <code>Node</code> to pretty print.
	 * @param config the <code>XMLPrettyPrinterConfiguration</code> object used
	 * for the pretty printing.
	 * @param buffer the <code>StringBuffer</code> used to accumulate the pretty
	 * printing of the <code>Node</code>.
	 * @param level the indentation level of the pretty printing.
	 * @see fr.inria.gemo.util.XMLPrettyPrinterConfiguration
	 */
	//	static public void NodeToString(
	//		Node node,
	//		XMLPrettyPrinterConfiguration config,
	//		StringBuffer buffer,
	//		int level) {
	static public void Serialize(
		Node node,
		OutputFormat outputFormat,
		Writer writer,
		int level) throws IOException {
		short type = node.getNodeType();
		switch (type) {
			case Node.ATTRIBUTE_NODE :
				Attr attr = (Attr) node;
				// Name is the namespace and the local name.
				// For example: axml:origin
				writer.write(attr.getName());
				writer.write("=\"");
				writer.write(EscapeText(attr.getValue()));
				writer.write('"');
				break;
			case Node.CDATA_SECTION_NODE :
				writer.write(outputFormat.getLineSeparator());
				if (outputFormat.getIndent())
					writer.write(
						StringUtils.CreateRepeatedString(
							outputFormat.getIndentString(),
							level));
				writer.write(CDATA_START);
				writer.write(node.getNodeValue());
				writer.write(CDATA_END);
				break;
			case Node.COMMENT_NODE :
				writer.write(outputFormat.getLineSeparator());
				if (outputFormat.getIndent())
					writer.write(
						StringUtils.CreateRepeatedString(
							outputFormat.getIndentString(),
							level));
				writer.write(COMMENT_START);
				writer.write(node.getNodeValue());
				writer.write(COMMENT_END);
				break;
			case Node.DOCUMENT_FRAGMENT_NODE :
				Serialize(
					node.getChildNodes(),
					outputFormat,
					writer,
					level + 1);
				break;
			case Node.DOCUMENT_NODE :
				writer.write(XML_HEADER_START);
				writer.write(" ");
				writer.write(XML_HEADER_VERSION);
				writer.write("=\"");
				writer.write(outputFormat.getXMLVersion());
				writer.write("\" ");
				writer.write(XML_HEADER_ENCODING);
				writer.write("=\"");
				writer.write(outputFormat.getXMLEncoding());
				writer.write("\" ");
				writer.write(XML_HEADER_STANDALONE);
				writer.write("=\"");
				writer.write(outputFormat.getXMLStandalone());
				writer.write("\" ");
				writer.write(XML_HEADER_END);
				Serialize(node.getChildNodes(), outputFormat, writer, level);
				break;
			case Node.DOCUMENT_TYPE_NODE :
				writer.write(outputFormat.getLineSeparator());
				if (outputFormat.getIndent())
					writer.write(
						StringUtils.CreateRepeatedString(
							outputFormat.getIndentString(),
							level));
				writer.write(DOCTYPE_START);
				writer.write(' ');
				DocumentType docType = (DocumentType) node;
				writer.write(docType.getName());
				writer.write(' ');
				// processing the external ID
				if (docType.getPublicId() != null) {
					writer.write(DOCTYPE_PUBLIC_ID);
					writer.write(" \"");
					writer.write(docType.getPublicId());
					writer.write("\" ");
				} else {
					writer.write(DOCTYPE_SYSTEM_ID);
					writer.write(' ');
				}
				writer.write('"');
				writer.write(docType.getSystemId());
				writer.write('"');

				// We will see later on what to do exactly !!!!!!!!!!!
				// TODO still missing some stuff, see http://www.w3.org/TR/REC-xml#dt-doctype
				// and http://www.w3.org/TR/2000/REC-DOM-Level-2-Core-20001113/core.html#ID-412266927
				writer.write(DOCTYPE_END);
				break;
			case Node.ELEMENT_NODE :
				writer.write(outputFormat.getLineSeparator());
				if (outputFormat.getIndent())
					writer.write(
						StringUtils.CreateRepeatedString(
							outputFormat.getIndentString(),
							level));
				Element element = (Element) node;
				// TagName is the namespace and the local name.
				// For example: axml:sc
				String name = element.getTagName();
				writer.write('<');
				writer.write(name);

				NamedNodeMap attrs = element.getAttributes();
				int nbAttrs = attrs.getLength();
				Attr a = null;
				for (int i = 0; i < nbAttrs; i++) {
					a = (Attr) attrs.item(i);
					writer.write(' ');
					Serialize(a, outputFormat, writer, level);
				}
				if (element.hasChildNodes()) {
					writer.write('>');
					Serialize(
						element.getChildNodes(),
						outputFormat,
						writer,
						level + 1);
					writer.write(outputFormat.getLineSeparator());
					if (outputFormat.getIndent())
						writer.write(
							StringUtils.CreateRepeatedString(
								outputFormat.getIndentString(),
								level));
					writer.write("</");
					writer.write(name);
					writer.write('>');
				} else {
					writer.write("/>");
				}
				break;
			case Node.ENTITY_NODE :
				// We will see later on what to do exactly
				break;
			case Node.ENTITY_REFERENCE_NODE :
				// We will see later on what to do exactly
				break;
			case Node.NOTATION_NODE :
				// We will see later on what to do exactly
				break;
			case Node.PROCESSING_INSTRUCTION_NODE :
				writer.write(outputFormat.getLineSeparator());
				if (outputFormat.getIndent())
					writer.write(
						StringUtils.CreateRepeatedString(
							outputFormat.getIndentString(),
							level));
				writer.write(PI_START);
				writer.write(node.getNodeName());
				writer.write(' ');
				// PIs are not part of the document's character data
				// so we don't need to escape its content.
				writer.write(node.getNodeValue());
				writer.write(PI_END);
				break;
			case Node.TEXT_NODE :
				if (outputFormat.getDiscardEmptyTextNodes()
					&& StringUtils.EMPTY_STRING.equals(
						node.getNodeValue().trim()))
					break;
				//				buffer.append(config.newline);
				//				if (config.indent)
				//					buffer.append(
				//						StringUtils.CreateRepeatedString(config.spaces, level));
				writer.write(EscapeText(node.getNodeValue()));
				break;
		}
	}

	/**
	 * Pretty prints an XML <code>NodeList</code> to a <code>StringBuffer</code> object.
	 * As the printing is done on the <code>Node</code>s and their children (if any), the
	 * <code>StringBuffer</code> buffer is used to accumulate the printing during the
	 * on-going process.
	 * @param nodes the <code>NodeList</code> to pretty print.
	 * @param config the <code>XMLPrettyPrinterConfiguration</code> object used
	 * for the pretty printing.
	 * @param buffer the <code>StringBuffer</code> used to accumulate the pretty
	 * printing of the <code>NodeList</code>.
	 * @param level the indentation level of the pretty printing.
	 * @see fr.inria.gemo.util.XMLPrettyPrinterConfiguration
	 */
	//	static public void ChildrenToString(
	//		NodeList nodes,
	//		XMLPrettyPrinterConfiguration config,
	//		StringBuffer buffer,
	//		int level) {
	static public void Serialize(
		NodeList nodes,
		OutputFormat outputFormat,
		Writer writer,
		int level) throws IOException {
		if (nodes.getLength() == 0)
			return;
		int nbChildren = nodes.getLength();
		for (int i = 0; i < nbChildren; i++) {
			Serialize(nodes.item(i), outputFormat, writer, level);
		}
	}

	/**
	 * Escapes the following characters <pre>& , ' , " , < , ></pre> from a
	 * <code>String</code> object. These characters are replaced by their
	 * predefined entities.
	 * @param s the <code>String</code> object from which to escape the
	 * characters.
	 * @return a <code>String</code> with the escaped characters replaced by
	 * their predefined entities.
	 */
	static public String EscapeText(String s) {
		StringBuffer buffer = new StringBuffer();
		char[] chars = s.toCharArray();
		int length = chars.length;
		for (int i = 0; i < length; i++) {
			switch (chars[i]) {
				case '&' :
					buffer.append(AMP);
					break;
				case '\'' :
					buffer.append(APOS);
					break;
				case '"' :
					buffer.append(QUOT);
					break;
				case '<' :
					buffer.append(LT);
					break;
				case '>' :
					buffer.append(GT);
					break;
				default :
					buffer.append(chars[i]);
					break;
			}
		}
		return buffer.toString();
	}
}
