package fr.inria.gemo.util.xml;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

import java.util.logging.Logger;

import org.w3c.dom.DocumentType;
import org.w3c.dom.Node;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Attr;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.w3c.dom.Entity;
import org.w3c.dom.Notation;
import org.w3c.dom.ProcessingInstruction;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Comment;

import fr.inria.gemo.util.StringUtils;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;

/**
 * <p>
 * This class provides utility functions for XML documents.
 * </p>
 * @author Jerome Baumgarten jerome.baumgarten@inria.fr
 * @version 1.0
 * 13/04/2003 : corrected a bug in EqualLocalNameAndNamespaceURI(Node, String, String).
 * It now uses the StringUtils.Equal(String, String) function to compare the Strings. Thus
 * it doesn't throw a NullPointerException if the namespace URI is null for example.
 * 21/01/2003 : added static methods for <code>Text</code> node nesting into
 * <code>Element</code>. Jerome Baumgarten.
 * 11/03/2003 : added static functions to compare DOM objects. Still need some deep
 * testing and the implementation for <code>DOCUMENT_FRAGMENT_NODE</code>.
 * Jerome Baumgarten.
 */
public class XMLUtils {

	static private Logger LOGGER = Logger.getLogger(XMLUtils.class.getName());

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	*
	* Constants.
	*
	* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

	/* Constants for the import attributes methods. */
	/**
	 * You don't want to have any duplicate attributes. Attributes are
	 * considered duplicate if they have same namespace URI and same local name.
	 */
	static public final short DUPLICATE_NO = 1;
	/**
	 * You allow duplicate attributes only they are equal, i.e. same
	 * namespace URI, same local name and same value.
	 */
	static public final short DUPLICATE_ONLY_IF_EQUAL = 2;
	/**
	 * If you want to overwrite the attributes of the destination <code>Element</code>
	 * with those from the source <code>Element</code>.
	 */
	static public final short DUPLICATE_OVERWRITE_DESTINATION = 3;
	/**
	 * If you want to keep the attributes of the destination <code>Element</code> and
	 * not import the duplicate ones from the source <code>Element</code>.
	 */
	static public final short DUPLICATE_KEEP_DESTINATION = 4;

	static private final DocumentBuilderFactory docBuilderFactory =
		DocumentBuilderFactory.newInstance();
	static private final DocumentBuilderFactory docBuilderFactoryNS =
		DocumentBuilderFactory.newInstance();
	static {
		XMLUtils.docBuilderFactoryNS.setNamespaceAware(true);
	}

	/**
	 * Tests if the <code>Node</code> n has the specified namespace URI and local name.
	 * @param node the node to test.
	 * @param localName the local name you want to compare to.
	 * @param namespaceURI the namespace URI you want to compare to.
	 * @return returns <code>true</code> if the <code>Node</code> n has for local name
	 * and namespace URI <code>localName</code> and <code>namespaceURI</code>,
	 * <code>false</code> otherwise.
	 */
	static public boolean EqualLocalNameAndNamespaceURI(
		Node node,
		String localName,
		String namespaceURI) {
		return StringUtils.Equal(localName, node.getLocalName())
			&& StringUtils.Equal(namespaceURI, node.getNamespaceURI());
	}

	/* ****************************
	 * The following method compares the LocalName and the namespace URI
	 * of the two elements.
	 * Return true if they are the same.
	 * ****************************/
	//	static public boolean equalLocalNameAndNameSpaceURI(Element e1, Element e2) {
	//		return fr.inria.gemo.util.StringUtils.equal(
	//			e1.getLocalName(),
	//			e2.getLocalName())
	//			&& fr.inria.gemo.util.StringUtils.equal(
	//				e1.getNamespaceURI(),
	//				e2.getNamespaceURI());
	//	}
	/**
	 * Tests if two <code>Node</code>s have the same namespace URI and same local name.
	 * @param node1 the first <code>Node</code> to test with.
	 * @param node2 the second <code>Node</code> to test with.
	 * @return <code>true</code> if the two <code>Node</code>s have the
	 * same namespace URI and the same local name, <code>false</code> otherwise.
	 */
	static public boolean EqualLocalNameAndNamespaceURI(
		Node node1,
		Node node2) {
		return fr.inria.gemo.util.StringUtils.Equal(
			node1.getLocalName(),
			node2.getLocalName())
			&& fr.inria.gemo.util.StringUtils.Equal(
				node1.getNamespaceURI(),
				node2.getNamespaceURI());
	}

	/**
	 * Imports the attributes from one <code>Element</code> to another one. You
	 * can specify the policy you want to apply for duplicate attributes. Depending on
	 * the policy, if the import fails, the <code>Attr</code> on which it failed will be
	 * returned. Please note that if the policy is <code>DUPLICATE_NO</code> or
	 * <code>DUPLICATE_ONLY_IF_EQUAL</code>, it will first check that the import
	 * is possible before actually doing it. This means that if the import cannot be
	 * performed, the destination <code>Element</code> will not be modified.
	 * @param destination the destination <code>Element</code> to which you
	 * want to import the attributes.
	 * @param source the source <code>Element</code> from which you want to
	 * import the attributes.
	 * @param policy the policy to apply to the import process regarding duplicate
	 * attributes, can be:
	 * <ul>
	 * <li>DUPLICATE_NO</li>
	 * <li>DUPLICATE_ONLY_IF_EQUAL</li>
	 * <li>DUPLICATE_OVERWRITE_DESTINATION</li>
	 * <li>DUPLICATE_KEEP_DESTINATION</li>
	 * </ul>
	 * @return <code>null</code> if the import of the attributes succeed, or the
	 * <code>Attr</code> on which it fails otherwise.
	 */
	static public Attr ImportAttributes(
		Element destination,
		Element source,
		short policy) {
		if (!source.hasAttributes()) // nothing to do.
			return null;
		if (!destination.hasAttributes()) {
			// we don't need to take in consideration the policy,
			// it just does an import.
			ImportAttributes(destination, source);
			return null;
		}
		Attr error = null;
		// DO NOT MOVE the IF AROUND !!
		// If the policy is DUPLICATE_NO or DUPLICATE_ONLY_IF_EQUAL then
		// we first need to check there's no conflict before actually doing the import.
		// If the policy cannot be respected, then it will return the Attr on which
		// it failed and do not the import at all, so the destination Element is not
		// modified.
		if ((policy == XMLUtils.DUPLICATE_NO)
			|| (policy == XMLUtils.DUPLICATE_ONLY_IF_EQUAL)) {
			error = CheckForDuplicateAttributes(destination, source, policy);
			if (error != null)
				return error;
		}
		switch (policy) {
			//			case XMLUtils.DUPLICATE_OVERWRITE_DESTINATION :
			//				ImportAttributesOverwriteDestination(destination, source);
			//				return null;
			case XMLUtils.DUPLICATE_KEEP_DESTINATION :
				ImportAttributesKeepDestination(destination, source);
				return null;
			case XMLUtils.DUPLICATE_OVERWRITE_DESTINATION :
			case XMLUtils.DUPLICATE_NO :
			case XMLUtils.DUPLICATE_ONLY_IF_EQUAL :
				ImportAttributes(destination, source);
				return null;
		}
		return null;
	}

	/**
	 * Imports the attributes from one <code>Element</code> to another one.
	 * If an <code>Attr</code> from the source <code>Element</code> is already
	 * present in the destination <code>Element</code>, the one from destination
	 * is replaced by the new one (the one from source). Two <code>Attr</code>s
	 * are considered the duplicate if they have the same namespace URI and the same
	 * local name.
	 * @param destination the destination <code>Element</code> to which you
	 * want to import the attributes.
	 * @param source the source <code>Element</code> from which you want to
	 * import the attributes.
	 */
	static public void ImportAttributes(Element destination, Element source) {
		NamedNodeMap attributes = source.getAttributes();
		int nbAttributes = attributes.getLength();
		for (int i = 0; i < nbAttributes; i++) {
			// cloning an attribute does not require a deep clone, at least I think so
			destination.setAttributeNodeNS(
				(Attr) attributes.item(i).cloneNode(false));
		}
	}

	/**
	 * Imports the attributes from one <code>Element</code> to another one.
	 * If an <code>Attr</code> from the source <code>Element</code> is already
	 * present in the destination <code>Element</code>, the one from destination
	 * is kept. Two <code>Attr</code>s are considered duplicate if they have the
	 * same namespace URI and the same local name.
	 * @param destination the destination <code>Element</code> to which you
	 * want to import the attributes.
	 * @param source the source <code>Element</code> from which you want to
	 * import the attributes.
	 */
	static protected void ImportAttributesKeepDestination(
		Element destination,
		Element source) {
		NamedNodeMap attributes = source.getAttributes();
		int nbAttributes = attributes.getLength();
		Attr attr = null;
		for (int i = 0; i < nbAttributes; i++) {
			attr = (Attr) attributes.item(i).cloneNode(false);
			if (destination
				.getAttributeNodeNS(attr.getNamespaceURI(), attr.getLocalName())
				!= null) // Keep the already existing Attr in destination.
				continue;
			else
				destination.setAttributeNodeNS(attr);
		}
	}

	/**
	 * Checks if two <code>Element</code>s have duplicate attributes. Two attributes
	 * are considered duplicate if they have the same namespace URI and the same local name.
	 * You can relax the test a little bit by using the policy. You can allow duplicate attributes
	 * if they have the same value as well, or totally reject duplicate attributes.
	 * @param e1 the first <code>Element</code> to check.
	 * @param e2 the second <code>Element</code> to check.
	 * @param policy the policy to apply to the check process regarding duplicate
	 * attributes, can be:
	 * <ul>
	 * <li>DUPLICATE_NO</li>
	 * <li>DUPLICATE_ONLY_IF_EQUAL</li>
	 * </ul>
	 * @return <code>null</code> if there is no duplicate attributes, or the duplicate
	 * <code>Attr</code> otherwise. This one can be either from <code>e1</code>
	 * or <code>e2</code>.
	 */
	static public Attr CheckForDuplicateAttributes(
		Element e1,
		Element e2,
		short policy) {
		if (!e2.hasAttributes() || !e1.hasAttributes())
			return null; // no duplicates

		Element element = null;
		NamedNodeMap attributes = null;

		// we iterate on the smallest set of attributes
		// the following is to get the smallest set.
		NamedNodeMap attrs1 = e1.getAttributes();
		NamedNodeMap attrs2 = e2.getAttributes();
		if (attrs1.getLength() < attrs2.getLength()) {
			attributes = attrs1; // attributes are from element e1
			element = e2;
		} else {
			attributes = attrs2; // attributes are from element e2
			element = e1;
		}
		// attrs1 = attrs2 = null; // not useful, will be gc'ed afterwards anyway. These are local variables.

		Node node = null;
		Attr attr1 = null;
		Attr attr2 = null;
		int nbAttributes = attributes.getLength();

		for (int i = 0; i < nbAttributes; i++) {
			node = attributes.item(i);
			attr1 = (Attr) node;
			attr2 =
				element.getAttributeNodeNS(
					attr1.getNamespaceURI(),
					attr1.getLocalName());
			if (attr2 == null)
				continue;
			switch (policy) {
				case XMLUtils.DUPLICATE_NO :
					return attr2;
				case XMLUtils.DUPLICATE_ONLY_IF_EQUAL :
					if (!StringUtils.Equal(attr1.getValue(), attr2.getValue()))
						return attr2;
					break;
			}
		}

		return null;
	}

	/** Apply the <code>Text</code> nodes nesting to the <code>Node</code>
	 * given as parameter. This can be a deep process. This means it will go deep into
	 * the children of the <code>Node</code> looking for <code>Text</code> nodes.
	 * @param node the <code>Node</code> to treat.
	 * @param tagName the name of the <code>Element</code> created to nest the
	 * <code>Text</code> node.
	 * @param namespaceURI the namespace URI of the <code>Element</code> created to nest the
	 * @param deep indicates if you want to do it deep into the <code>Node</code>s
	 * or only for the top <code>Node</code>s
	 * <code>Text</code> node.
	 */
	static public void TextNodeNesting(
		Node node,
		String tagName,
		String namespaceURI,
		boolean deep) {
		short type = node.getNodeType();
		switch (type) {
			case Node.ATTRIBUTE_NODE :
			case Node.CDATA_SECTION_NODE :
			case Node.COMMENT_NODE :
			case Node.DOCUMENT_FRAGMENT_NODE :
			case Node.DOCUMENT_TYPE_NODE :
			case Node.ENTITY_NODE :
			case Node.ENTITY_REFERENCE_NODE :
			case Node.NOTATION_NODE :
			case Node.PROCESSING_INSTRUCTION_NODE :
				break;
			case Node.DOCUMENT_NODE :
				// TextNodeNesting(node.getChildNodes(),tagName,namespaceURI);
				// break;
			case Node.ELEMENT_NODE :
				if (deep)
					TextNodeNesting(
						node.getChildNodes(),
						tagName,
						namespaceURI,
						deep);
				break;
			case Node.TEXT_NODE :
				TextNodeNesting((Text) node, tagName, namespaceURI);
				break;
		}
	}

	/**
	 * Nest a <code>Text</code> node into an <code>Element</code>. The created
	 * <code>Element</code> will be inserted into the XML tree and replace the
	 * <code>Text</code> node.
	 * @param text the <code>Text</code> node to nest.
	 * @param tagName the name of the <code>Element</code> created to nest the
	 * <code>Text</code> node.
	 * @param namespaceURI the namespace URI of the <code>Element</code> created to nest the
	 * <code>Text</code> node.
	 * @return the <code>Element</code> nesting the <code>Text</code> node.
	 */
	static public Element TextNodeNesting(
		Text text,
		String tagName,
		String namespaceURI) {
		Element element = null;
		if (!StringUtils.IsEmpty(namespaceURI)) {
			element =
				text.getOwnerDocument().createElementNS(namespaceURI, tagName);
		} else {
			element = text.getOwnerDocument().createElement(tagName);
		}
		element.appendChild(text.cloneNode(false));
		Node parent = text.getParentNode();
		parent.replaceChild(element, text);
		return element;
	}

	/** Apply the <code>Text</code> nodes nesting to the nodes of the <code>NodeList</code>
	 * given as parameter.
	 * @param nodes the <code>NodeList</code> to treat.
	 * @param tagName the name of the <code>Element</code> created to nest the
	 * <code>Text</code> node.
	 * @param namespaceURI the namespace URI of the <code>Element</code> created to nest the
	 * <code>Text</code> node.
	 */
	static public void TextNodeNesting(
		NodeList nodes,
		String tagName,
		String namespaceURI,
		boolean deep) {
		int nbNodes = nodes.getLength();
		for (int i = 0; i < nbNodes; i++)
			TextNodeNesting(nodes.item(i), tagName, namespaceURI, deep);
	}

	/** Returns the first direct child <code>Element</code> that has as local name
	 * <code>localName</code> and as namespace URI <code>namespaceURI</code>. If
	 * there's any it returns <code>null</code>.
	 * @param element the parent <code>Element</code> on which to do the search.
	 * @param namespaceURI the namespace URI of the <code>Element</code> you're
	 * looking for.
	 * @param localName the local name of the <code>Element</code> to find.
	 * @return the first direct child <code>Element</code> if any, <code>null</code>
	 * otherwise.
	 */
	static public Element GetFirstChildElementByTagNameNS(
		Element element,
		String namespaceURI,
		String localName) {
		if (!element.hasChildNodes()) {
			return null;
		}
		NodeList children = element.getChildNodes();
		int l = children.getLength();
		Node n = null;
		for (int i = 0; i < l; i++) {
			n = children.item(i);
			if (n.getNodeType() == Node.ELEMENT_NODE) {
				if (XMLUtils
					.EqualLocalNameAndNamespaceURI(
						n,
						localName,
						namespaceURI)) {
					return (Element) n;
				}
			} else {
				continue;
			}
		}
		return null;
	}

	static public Node GetFirstChildNodeByType(Node node, short nodeType) {
		if (node == null)
			return null;
		return GetFirstNodeByType(node.getChildNodes(), nodeType);
	}

	static public Node GetFirstNodeByType(NodeList nodes, short nodeType) {
		int nbNodes = nodes.getLength();
		for (int i = 0; i < nbNodes; i++) {
			Node node = nodes.item(i);
			if (node.getNodeType() == nodeType)
				return node;
		}
		return null;
	}

	static public boolean AreAllChildNodes(Node node, short nodeType) {
		if (node.hasChildNodes()) {
			NodeList childNodes = node.getChildNodes();
			int nbChildNodes = childNodes.getLength();
			for (int i=0 ; i<nbChildNodes ; i++) {
				Node child = childNodes.item(i);
				if (child.getNodeType() != nodeType)
					return false;
			}
		}
		return true;
	}

	static public DocumentBuilderFactory GetDocumentBuilderFactory(boolean namespaceAware) {
		if (namespaceAware) {
			return XMLUtils.docBuilderFactoryNS;
		} else {
			return XMLUtils.docBuilderFactory;
		}
	}

	static public DocumentBuilder GetDocumentBuilder(boolean namespaceAware)
		throws ParserConfigurationException {
		/** We build a new DocumentBuilder for every request because of the following recommendation :
		 * An implementation of DocumentBuilder is NOT guaranteed to behave as per
		 * the specification if it is used concurrently by two or more threads.
		 * It is recommended to have one instance of the DocumentBuilder per thread
		 * or it is upto the application to make sure about the use of DocumentBuilder from more than one thread.
		 */
		if (namespaceAware) {
			return XMLUtils.docBuilderFactoryNS.newDocumentBuilder();
		} else {
			return XMLUtils.docBuilderFactory.newDocumentBuilder();
		}
	}

	static public Document CreateNewDocument(boolean namespaceAware)
		throws ParserConfigurationException {
		return GetDocumentBuilder(namespaceAware).newDocument();
	}

	/** All these Equal functions won't be useful anymore once the DOM Level 3
	 * API is out and Xerces provides a standard implementation (so far its implementation
	 * is experimental and to get access to it you need to compile the Xerces sources
	 * and explicitly ask for DOM 3 support).
	 * @param node1
	 * @param node2
	 * @param deep
	 * @return boolean
	 */
	static public boolean Equal(Node node1, Node node2, boolean deep) {
		if ((node1 == null) && (node2 == null)) {
			return true;
		}

		if ((node1 == null) || (node2 == null)) {
			return false;
		}

		short type1 = node1.getNodeType();
		short type2 = node2.getNodeType();

		if (type1 != type2) {
			return false;
		}

		switch (type1) {
			case Node.ATTRIBUTE_NODE :
				return Equal((Attr) node1, (Attr) node2);
			case Node.CDATA_SECTION_NODE :
				return Equal((CDATASection) node1, (CDATASection) node2);
			case Node.COMMENT_NODE :
				return Equal((Comment) node1, (Comment) node2);
			case Node.DOCUMENT_FRAGMENT_NODE :
				// TODO : implement the DOCUMENT_FRAGMENT_NODE comparison
				throw new RuntimeException("Not yet implemented...");
			case Node.DOCUMENT_TYPE_NODE :
				return Equal((DocumentType) node1, (DocumentType) node2);
			case Node.ENTITY_NODE :
				return Equal((Entity) node1, (Entity) node2);
			case Node.ENTITY_REFERENCE_NODE :
				return StringUtils.Equal(
					node1.getNodeName(),
					node2.getNodeName());
			case Node.NOTATION_NODE :
				return Equal((Notation) node1, (Notation) node2);
			case Node.PROCESSING_INSTRUCTION_NODE :
				return Equal(
					(ProcessingInstruction) node1,
					(ProcessingInstruction) node2);
			case Node.DOCUMENT_NODE :
				return Equal(
					node1.getChildNodes(),
					node2.getChildNodes(),
					deep);
			case Node.ELEMENT_NODE :
				return Equal((Element) node1, (Element) node2, deep);
			case Node.TEXT_NODE :
				return Equal((Text) node1, (Text) node2);
		}

		return false;
	}

	static public boolean Equal(CDATASection cdata1, CDATASection cdata2) {
		if ((cdata1 == null) && (cdata2 == null))
			return true;
		if ((cdata1 == null) || (cdata2 == null))
			return false;
		return StringUtils.Equal(cdata1.getNodeValue(), cdata2.getNodeValue());
	}

	static public boolean Equal(Comment comment1, Comment comment2) {
		if ((comment1 == null) && (comment2 == null))
			return true;
		if ((comment1 == null) || (comment2 == null))
			return false;
		return StringUtils.Equal(
			comment1.getNodeValue(),
			comment2.getNodeValue());
	}

	static public boolean Equal(Text text1, Text text2) {
		if ((text1 == null) && (text2 == null))
			return true;
		if ((text1 == null) || (text2 == null))
			return false;
		return StringUtils.Equal(
			text1.getNodeValue().trim(),
			text2.getNodeValue().trim());
	}

	static public boolean Equal(
		Collection nodes1,
		Collection nodes2,
		boolean deep) {
		if (nodes1.size() != nodes2.size())
			return false;
		Iterator iteratorNodes1 = nodes1.iterator();
		Iterator iteratorNodes2 = nodes2.iterator();
		while (iteratorNodes1.hasNext()) {
			Node node1 = (Node) iteratorNodes1.next();
			Node node2 = (Node) iteratorNodes2.next();
			if (!Equal(node1, node2, deep))
				return false;
		}
		return true;
	}

	static public boolean Equal(
		NodeList nodes1,
		NodeList nodes2,
		boolean deep) {
		if (nodes1.getLength() != nodes2.getLength())
			return false;
		int length = nodes1.getLength();
		for (int i = 0; i < length; i++) {
			if (!Equal(nodes1.item(i), nodes2.item(i), deep))
				return false;
		}
		return true;
	}

	static public boolean Equal(Attr attr1, Attr attr2) {
		// LOGGER.finest("Equal(" + ((attr1 == null) ? "NULL" : DOMSerializer.Serialize(attr1)) + "," + ((attr2 == null) ? "NULL" : DOMSerializer.Serialize(attr2)) + ")");
		if ((attr1 == null) && (attr2 == null)) {
			// LOGGER.finest("Equal(Attr,Attr) : attributes are both null : returns true");
			return true;
		}
		if ((attr1 == null) || (attr2 == null)) {
			// LOGGER.finest("Equal(Attr,Attr) : only one of the attribute is null : returns false");
			return false;
		}
		return EqualLocalNameAndNamespaceURI(attr1, attr2)
			&& StringUtils.Equal(attr1.getNodeValue(), attr2.getNodeValue());
	}

	static public boolean Equal(Entity entity1, Entity entity2) {
		if ((entity1 == null) && (entity2 == null))
			return true;
		if ((entity1 == null) || (entity2 == null))
			return false;
		// TODO : check the code...
		// If the entities are unparsed, we need to compare the notation name
		if (!StringUtils
			.Equal(entity1.getNotationName(), entity2.getNotationName())) {
			return false;
		}
		// Try to compare the SystemId or the PublicId
		if (!StringUtils.Equal(entity1.getSystemId(), entity2.getSystemId())) {
			return false;
		}
		if (!StringUtils.Equal(entity1.getPublicId(), entity2.getPublicId())) {
			return false;
		}
		// If the entites are parsed, we need to compare the children
		NodeList children1 = entity1.getChildNodes();
		NodeList children2 = entity2.getChildNodes();
		if (children1.getLength() == children2.getLength()) {
			for (int i = 0; i < children1.getLength(); i++) {
				if (!Equal(children1.item(i), children2.item(2), true))
					return false;
			}
		}
		return true;
	}

	static public boolean Equal(Notation notation1, Notation notation2) {
		if ((notation1 == null) && (notation2 == null))
			return true;
		if ((notation1 == null) || (notation2 == null))
			return false;
		if (!StringUtils
			.Equal(notation1.getNodeValue(), notation2.getNodeValue())) {
			return false;
		}
		if (!StringUtils
			.Equal(notation1.getSystemId(), notation2.getSystemId())) {
			return false;
		}
		if (!StringUtils
			.Equal(notation1.getPublicId(), notation2.getPublicId())) {
			return false;
		}
		return true;
	}

	static public boolean Equal(
		ProcessingInstruction pi1,
		ProcessingInstruction pi2) {
		if ((pi1 == null) && (pi2 == null))
			return true;
		if ((pi1 == null) || (pi2 == null))
			return false;
		return StringUtils.Equal(pi1.getNodeName(), pi2.getNodeName())
			&& StringUtils.Equal(pi1.getNodeValue(), pi2.getNodeValue());
	}

	static public boolean Equal(DocumentType dt1, DocumentType dt2) {
		if ((dt1 == null) && (dt2 == null))
			return true;
		if ((dt1 == null) || (dt2 == null))
			return false;
		return StringUtils.Equal(dt1.getPublicId(), dt2.getPublicId())
			&& StringUtils.Equal(dt1.getSystemId(), dt2.getSystemId());
	}

	static public boolean Equal(
		Element element1,
		Element element2,
		boolean deep) {
		// LOGGER.finest("Equal(Element,Element)");
		if ((element1 == null) && (element2 == null)) {
			// LOGGER.finest("Equal(Element,Element) : elements are both null : returns true");
			return true;
		}
		if ((element1 == null) || (element2 == null)) {
			// LOGGER.finest("Equal(Element,Element) : only one of the element is null : returns false");
			return false;
		}

		//		LOGGER.finest(
		//			"Element 1 : "
		//				+ ((element1.getNamespaceURI() != null)
		//					? "{" + element1.getNamespaceURI() + "}"
		//					: "")
		//				+ element1.getNodeName());
		//
		//			LOGGER.finest(
		//				"Element 2 : "
		//					+ ((element2.getNamespaceURI() != null)
		//						? "{" + element2.getNamespaceURI() + "}"
		//						: "")
		//					+ element2.getNodeName());

		// Check the local name and the namespace URI.
		if (!EqualLocalNameAndNamespaceURI(element1, element2)) {
			// LOGGER.finest("Equal(Element,Element) : elements have different local names and namespace URIs : returns false");
			return false;
		}

		// Check the attributes.
		NamedNodeMap attrs1 = element1.getAttributes();
		NamedNodeMap attrs2 = element2.getAttributes();
		if (attrs1.getLength() != attrs2.getLength()) {
			// LOGGER.finest("Equal(Element,Element) : elements have different number of attributes : returns false");
			return false;
		}
		Attr attr1, attr2;
		for (int i = 0; i < attrs1.getLength(); i++) {
			attr1 = (Attr) attrs1.item(i);
			//			LOGGER.finest(
			//				"Attribute 1 : "
			//					+ ((attr1.getNamespaceURI() != null)
			//						? "{" + attr1.getNamespaceURI() + "}"
			//						: "")
			//					+ attr1.getName() + "(localName=" + attr1.getLocalName() + ")");
			attr2 =
				(Attr) attrs2.getNamedItemNS(
					attr1.getNamespaceURI(),
					attr1.getLocalName());
			if (!Equal(attr1, attr2))
				return false;
		}

		// Check the children if deep is true
		if (deep)
			return Equal(
				element1.getChildNodes(),
				element2.getChildNodes(),
				true);

		return true;
	}

	static public void RemoveAllChildren(Node node) {
		Node child = node.getFirstChild();
		while (child != null) {
			node.removeChild(child);
			child = node.getFirstChild();
		}
	}

	static public void MoveChildren(
		Node destination,
		Node source) {
		Node moved = source.getFirstChild();
		while (moved != null) {
			// destination.appendChild(moved.cloneNode(deepClone));
			// source.removeChild(moved);
			destination.appendChild(moved);
			moved = source.getFirstChild();
		}
	}

	static public ArrayList NodeList2ArrayList(NodeList nodes) {
		int nbNodes = nodes.getLength();
		ArrayList results = new ArrayList(nbNodes);
		for (int i = 0; i < nbNodes; i++)
			results.add(i, nodes.item(i));
		return results;
	}
	
	// Treats attributes as children of Elements
	public static Node getParent(Node n) {
		Node parentNode;
		if (n.getNodeType() == Node.ATTRIBUTE_NODE) {
			parentNode = ((Attr) n).getOwnerElement();
		} else {
			parentNode = n.getParentNode();
		}
		return parentNode;
	}

	// Allows replacing any combination of Nodes
	public static void replaceChild(
		Node parentNode,
		Node newChild,
		Node oldChild) {

		if (oldChild.getNodeType() == Node.ATTRIBUTE_NODE) {
			Attr oldAttr = (Attr) oldChild;
			Element parentElement = (Element) parentNode;
			parentElement.removeAttributeNode(oldAttr);
			if (newChild==null) {
				return;
			}
			if (newChild.getNodeType() == Node.ATTRIBUTE_NODE) {
				Attr newAttr = (Attr) newChild;
				parentElement.setAttributeNode(newAttr);
			} else {
				parentElement.appendChild(newChild);
			}

		} else {
			if (newChild==null) {
				parentNode.removeChild(oldChild);
			}
			else if (newChild.getNodeType() == Node.ATTRIBUTE_NODE) {
				Attr newAttr = (Attr) newChild;
				Element parentElement = (Element) parentNode;
				parentElement.removeChild(oldChild);
				parentElement.setAttributeNode(newAttr);
			} else {
				parentNode.replaceChild(newChild, oldChild);
			}
		}
	}
	
	/**
	 *  returns vector of child elements with a given name, 
	 *  vector is empty if not found 
	 * @param parent            - current element
	 * @param childElementName  - name of elements to returns
	 */
	public static Vector getSubElements(Element parent, String childElementName){
		if (parent == null || childElementName == null){
			return null;
		}
		
		NodeList nl = parent.getChildNodes();
		Vector els  = new Vector();
		//int count   = 0;
		
		for (int j=0; j <  nl.getLength(); j++)
		{
			org.w3c.dom.Node childNode = nl.item(j);
			if (childNode.getNodeName() == childElementName) {
				els.add(childNode);
			}
		}
		return els;
	}
}